#include "CSACFile.h"
#include "MERLIN_Error.h"

using CSAC::MERLIN_Error;

void
CSACFile::close() 
throw (string)
{
  int status = 0;
	
  if ( fits_close_file(_fptr, &status) )
  {
    string error = string("Can not close FITS file ") + _filename + string(".\n");
    string error_status;
    process_status(status,error_status);
    error+=" The CFITSIO library reports: " + error_status; 
    throw error;
  }
  
  _fptr = 0;
  
  _data_type = "";
  
  _open = false;
  
}

CSACFile::~CSACFile()
{
  try
  {
    if(_open)
      close();
  }
  catch(string &ex)
  {
    // This is a fatal error, something is wrong and we can't fix it 
    // because we are in a destructor
    cerr << __FILE__ << ":" << __LINE__ << ": While trying to close FITS file, got the following exception\n" << ex << endl;
    cerr << "STOP! must exit application now since this error ocurred inside a destructor and this is not recorverable." << endl;
    exit(1);
  }
}

void
CSACFile::real_open() 
{
  int status = 0;
	
  if ( fits_open_file(&_fptr, _filename.c_str(), READONLY, &status) )
  {
    string error="Can not open FITS file ";
    error+=_filename;
    string error_status;
    process_status(status,error_status);
    error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
    throw error;
  }
	
  int hdunum;
  if ( fits_get_num_hdus(_fptr, &hdunum, &status) )
  {
    string error="Can not get total number of Header-Data Units in this FITS file ";
    error+=_filename;
    string error_status;
    process_status(status,error_status);
    error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
    throw error;
  }
	
  if(hdunum > 1)
      cout<<string("Warning!: There is more than one HDU in this FITS file")<<endl;
  cout<<string("Using the first HDU found")<<endl;; 
  
  int hdutype;
  if ( fits_get_hdu_type(_fptr, &hdutype, &status) )
  {
      string error="Can not get the type of the current of Header-Data Units in this FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
  }
  
  if(hdutype!=IMAGE_HDU)
      throw string("The HDU type found is not supported by this program. HDU must be type IMAGE_HDU");
  
  char value[FLEN_VALUE];	
  if(fits_read_keyword(_fptr,"BITPIX",value,NULL,&status)) 
  {
      string error="Can not get the value for keyword BITPIX ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
  }
  
  int bitpix=atoi(value);
  bool right_bitpix;
  string bitpix_analysis_message;
  switch(bitpix)
  {
      case BYTE_IMG:
          bitpix_analysis_message="The type of data for the image is not supported, found BYTE_IMG";
          right_bitpix=false;
          break;
      case SHORT_IMG:
          bitpix_analysis_message="The type of data for the image is not supported, found SHORT_IMG";
          _data_type = "SHORT_IMG";
          right_bitpix=true;
          break;
      case LONG_IMG:
          bitpix_analysis_message="The type of data for the image is not supported, found LONG_IMG";
          right_bitpix=false;
          break;
      case FLOAT_IMG:
          bitpix_analysis_message="The type of data for the image is not supported, found FLOAT_IMG";
          _data_type = "FLOAT_IMG";
          right_bitpix=true;
          break;
      case DOUBLE_IMG:
          bitpix_analysis_message="The type of data for the image is not supported, found DOUBLE_IMG";
          right_bitpix=false;
          break;
      default:
          bitpix_analysis_message="The returned bitpix type is unknown";
          right_bitpix=false;
          break;
  }   
  if(!right_bitpix)
      throw bitpix_analysis_message;
  
  if(fits_read_keyword(_fptr,"NAXIS",value,NULL,&status)) 
  {
      string error="Can not get the value for keyword NAXIS ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
  }
  
  int number_axes=atoi(value);
  
  if(number_axes != 3)
      throw string("There are not exactly 3 axis (y, lambda, type of stoke vector) in the file");
  
  long int *axes;
  axes = new long int [number_axes];
  int nfound;
  if(fits_read_keys_lng(_fptr, "NAXIS", 1, number_axes, axes, &nfound, &status) )
  {
      string error="Can not get the value for one of the keywords NAXIS1, NAXIS2 or NAXIS3 ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
  }
  
  if(nfound==0)
      throw string("The number of keywords with the root name NAXIS was not found");
  
  _number_slit_positions = axes[1];
  _lambda = axes[0];
  
  if (axes[2] != 4)
      throw string("The number of vectors for each y/lambda position is not 4");
  
  if(!fits_read_keyword(_fptr,"SPBSHFT",value,NULL,&status)) 
  {
      int spbshft = atoi(value);
      if (spbshft > 0)
      {
          switch (spbshft)
          {
              case 1:
                  _scale_I = true;
                  break;
              case 2:
                  _scale_V = true;
                  break;
              case 3:
                  _scale_Q = true;
                  _scale_U = true;
                  break;
              default:
                  throw string("Invalid value for keyword SPBSHFT");
                  break;
          }
      }
  }
  else
      cout<<"Keyword SPBSHFT not found"<<endl;
  
  _open = true;

  delete [] axes;
}

void 
CSACFile::show_header()
{
  int status = 0;
  int nkeys;
	
  if(fits_get_hdrspace(_fptr, &nkeys, NULL, &status))
  {
    string error="Can not read header from FITS file ";
    error+=_filename;
    string error_status;
    process_status(status,error_status);
    error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
    throw error;
  }
	
  char card[FLEN_CARD]; 
  for (int ii = 1; ii <= nkeys; ii++)  
  { 
    if (fits_read_record(_fptr, ii, card, &status))
    {
      string error="Can not read record from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    printf("%s\n", card);
  }
}

void 
CSACFile::get_I(int pos, vector<double> &data)
{
  if(pos > _number_slit_positions)
    throw string("Invalid slit position");
		
  long fpixel = 1 + ((pos - 1) * _lambda);
  float nullval=0;
  int anynull;
  int status=0;
  
  if (data.size() != _lambda)
    data.resize(_lambda, 0.0);
  
  if (_data_type == "FLOAT_IMG")
  {
	  float *my_data;
    my_data = new float [_lambda];
    if ( fits_read_img(_fptr,TFLOAT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(double)my_data[h];
        if (data[h] < 0)
            data[h] = data[h] + 65536;
        if (_scale_I)
            data[h] = 2. * data[h]; 
    }
	delete [] my_data;
  }
  else if (_data_type == "SHORT_IMG")
  {
	  short *my_data;
    my_data = new short [_lambda];
    if ( fits_read_img(_fptr,TSHORT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(double)my_data[h];
        if (data[h] < 0)
            data[h] = data[h] + 65536;
        if (_scale_I)
            data[h] = 2. * data[h]; 
    }
	delete [] my_data;
  }
  else
    MERLIN_Error::errorT(1,"ssiss",__FILE__,":",__LINE__,": Invalid data type for this FITS file, found type ",_data_type.c_str());
}

void 
CSACFile::get_I(int pos, float *data, const int &len)
{
  if(pos > _number_slit_positions)
    throw string("Invalid slit position");
  
  if (len != _lambda)
    throw string ("get_I error: size missmatch.");
  
  long fpixel = 1 + ((pos - 1) * _lambda);
  float nullval=0;
  int anynull;
  int status=0;
  
  if (_data_type == "FLOAT_IMG")
  {
    if ( fits_read_img(_fptr,TFLOAT, fpixel, _lambda, &nullval, data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        if (data[h] < 0)
            data[h] = data[h] + 65536;
        if (_scale_I)
            data[h] = 2. * data[h];
    }
  }
  else if (_data_type == "SHORT_IMG")
  {
	  short *my_data;
    my_data = new short [_lambda];
    if ( fits_read_img(_fptr,TSHORT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(float)my_data[h];
        if (data[h] < 0)
            data[h] = data[h] + 65536;
        if (_scale_I)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  }
  else
    MERLIN_Error::errorT(1,"ssiss",__FILE__,":",__LINE__,": Invalid data type for this FITS file, found type ",_data_type.c_str());
}

void 
CSACFile::get_Q(int pos, vector<double> &data)
{
  if(pos > _number_slit_positions)
    throw string("Invalid slit position");
	
  long fpixel = (1 + _number_slit_positions * _lambda) + ((pos - 1) * _lambda);
  float nullval=0;
  int anynull;
  int status=0;
  
  if (data.size() != _lambda)
    data.resize(_lambda, 0.0);
  
  if (_data_type == "FLOAT_IMG")
  {	
	  float *my_data;
    my_data = new float [_lambda];
    if ( fits_read_img(_fptr,TFLOAT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data Q from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(double)my_data[h];
        if (_scale_Q)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  }
  else if (_data_type == "SHORT_IMG")
  {
	  short *my_data;
    my_data = new short [_lambda];
    if ( fits_read_img(_fptr,TSHORT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(double)my_data[h];
        if (_scale_Q)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  }
  else
    MERLIN_Error::errorT(1,"ssiss",__FILE__,":",__LINE__,": Invalid data type for this FITS file, found type ",_data_type.c_str());
  
}

void 
CSACFile::get_Q(int pos, float *data, const int &len)
{
  if(pos > _number_slit_positions)
    throw string("Invalid slit position");
  
  if (len != _lambda)
    throw string ("get_Q error: size missmatch.");
  
  long fpixel = (1 + _number_slit_positions * _lambda) + ((pos - 1) * _lambda);
  float nullval=0;
  int anynull;
  int status=0;
  
  if (_data_type == "FLOAT_IMG")
  {
    if ( fits_read_img(_fptr,TFLOAT, fpixel, _lambda, &nullval, data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    if (_scale_Q)
    {
        for (int h=0; h < _lambda; h++)
            data[h] = 2. * data[h];
    }
  }
  else if (_data_type == "SHORT_IMG")
  {
	  short *my_data;
    my_data = new short [_lambda];
    if ( fits_read_img(_fptr,TSHORT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(float)my_data[h];
         if (_scale_Q)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  }
  else
    MERLIN_Error::errorT(1,"ssiss",__FILE__,":",__LINE__,": Invalid data type for this FITS file, found type ",_data_type.c_str());
}

void 
CSACFile::get_U(int pos, vector<double> &data)
{
  if(pos > _number_slit_positions)
    throw string("Invalid slit position");
		
  long fpixel = (1 + 2 * _number_slit_positions * _lambda) + ((pos - 1) * _lambda);
  float nullval=0;
  int anynull;
  int status=0;
  
  if (data.size() != _lambda)
    data.resize(_lambda, 0.0);
  
  if (_data_type == "FLOAT_IMG")
  {
	  float *my_data;
    my_data = new float [_lambda];
    if ( fits_read_img(_fptr,TFLOAT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data U from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(double)my_data[h];
        if (_scale_U)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  } 
  else if (_data_type == "SHORT_IMG")
  {
	  short *my_data;
    my_data = new short [_lambda];
    if ( fits_read_img(_fptr,TSHORT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(double)my_data[h];
        if (_scale_U)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  }
  else
    MERLIN_Error::errorT(1,"ssiss",__FILE__,":",__LINE__,": Invalid data type for this FITS file, found type ",_data_type.c_str());
}

void 
CSACFile::get_U(int pos, float *data, const int &len)
{
  if(pos > _number_slit_positions)
    throw string("Invalid slit position");
  
  if (len != _lambda)
    throw string ("get_U error: size missmatch.");
  
  long fpixel = (1 + 2 * _number_slit_positions * _lambda) + ((pos - 1) * _lambda);
  float nullval=0;
  int anynull;
  int status=0;
  
  if (_data_type == "FLOAT_IMG")
  {
    if ( fits_read_img(_fptr,TFLOAT, fpixel, _lambda, &nullval, data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    if (_scale_U)
    {
        for (int h=0; h < _lambda; h++)
            data[h] = 2. * data[h];
    }
  }
  else if (_data_type == "SHORT_IMG")
  {
	  short *my_data;
    my_data = new short [_lambda];
    if ( fits_read_img(_fptr,TSHORT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(float)my_data[h];
        if (_scale_U)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  }
  else
    MERLIN_Error::errorT(1,"ssiss",__FILE__,":",__LINE__,": Invalid data type for this FITS file, found type ",_data_type.c_str());
}

void 
CSACFile::get_V(int pos, vector<double> &data)
{
  
  if(pos > _number_slit_positions)
    throw string("Invalid slit position");
	
  long fpixel= (1 + 3 * _number_slit_positions * _lambda) + ((pos - 1) * _lambda);
  float nullval=0;
  int anynull;
  int status=0;
  
  if (data.size() != _lambda)
    data.resize(_lambda, 0.0);
  
  if (_data_type == "FLOAT_IMG")
  {
	  float *my_data;
    my_data = new float [_lambda];
    if ( fits_read_img(_fptr,TFLOAT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data V from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(double)my_data[h];
        if (_scale_V)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  }
  else if (_data_type == "SHORT_IMG")
  {
	  short *my_data;
    my_data = new short [_lambda];
    if ( fits_read_img(_fptr,TSHORT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(double)my_data[h];
        if (_scale_V)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  }
  else
    MERLIN_Error::errorT(1,"ssiss",__FILE__,":",__LINE__,": Invalid data type for this FITS file, found type ",_data_type.c_str());
}

void 
CSACFile::get_V(int pos, float *data, const int &len)
{
  if(pos > _number_slit_positions)
    throw string("Invalid slit position");
  
  if (len != _lambda)
    throw string ("get_V error: size missmatch.");
  
  long fpixel= (1 + 3 * _number_slit_positions * _lambda) + ((pos - 1) * _lambda);
  float nullval=0;
  int anynull;
  int status=0;
  
  if (_data_type == "FLOAT_IMG")
  {
    if ( fits_read_img(_fptr,TFLOAT, fpixel, _lambda, &nullval, data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    if (_scale_V)
    {
        for (int h=0; h < _lambda; h++)
            data[h] = 2. * data[h];
    }
  }
  else if (_data_type == "SHORT_IMG")
  {
	  short *my_data;
    my_data = new short [_lambda];
    if ( fits_read_img(_fptr,TSHORT, fpixel, _lambda, &nullval, my_data, &anynull, &status) )
    {
      string error="Can not read data I from FITS file ";
      error+=_filename;
      string error_status;
      process_status(status,error_status);
      error+=" \nThe CFITSIO library reports: "+error_status+"\n"; 
      throw error;
    }
    for (int h=0; h < _lambda; h++)
    {
        data[h]=(float)my_data[h];
        if (_scale_V)
            data[h] = 2. * data[h];
    }
	delete [] my_data;
  }
  else
    MERLIN_Error::errorT(1,"ssiss",__FILE__,":",__LINE__,": Invalid data type for this FITS file, found type ",_data_type.c_str());
}

