/* ============================================================================
 * Language: C++ 
 * File: merlin_init.cc
 * Package Name: Milne Eddington gRid Linear Inversion Network (MERLIN).
 * Sub-package: Stand Alone Single Core Initialization Client. 
 *
 * Created by Jose Garcia <jgarcia@ucar.edu> on 07/16/07.
 *  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact University Corporation for Atmospheric Research at
 * 3080 Center Green Drive, Boulder, CO 80301

 * (c) COPYRIGHT University Corporation for Atmostpheric Research 2004-2006
 * Please read the full copyright statement in the file COPYRIGHT_UCAR.
 *
 =========================================================================== */


#include <math.h>

#include <iostream>
#include <fstream>
#include <vector>
using namespace std;

#include "config.h"
#include "MERLIN.h"
#include "MERLIN_Init.h"

// =============================================================================
// *                                                                           *
// * Interface Declaration to the merlin_initialization library that is        *
// * mostly in Fortran 77 and Fortran 90                                       *
// *                                                                           *
// =============================================================================

#define PIKAIA_FC FC_FUNC (pikaia, PIKAIA)
#define RNINIT_FC FC_FUNC (rninit, RNINIT)
#define RUFFAZM_FC FC_FUNC (ruffazm, RUFFAZM)
#define RUFFCSN_FC FC_FUNC (ruffcsn, RUFFCSN)
#define RUFFFLD_FC FC_FUNC (rufffld, RUFFFLD)
#define INVERT_FIELD_FC FC_FUNC_ (invert_field, INVERT_FIELD)

typedef float (*pikaia_call_back) (int*, float*);

extern "C"
{
	// f = ff(n,x), where x is a real parameter array of length n.
	float my_ff(int *n, float* x);
	void PIKAIA_FC(pikaia_call_back, int *n, float* ctrl, float* x, float *f, int *status);
	// Random Number Generator Initialization to seed pikaia
	void RNINIT_FC( int* seed );
	// Ruff initialization for the azimuth
	void RUFFAZM_FC(int *len, float *obsq, float *obsu, int *limblu, 
			float *cen, int *limred, float *dvmin, float *azm);
	// Ruff initialization for the cosine of the inclination of the
	// magnetic field vector
	void RUFFCSN_FC(int *len, float *obsq, float *obsu, float *obsv, 
			int *limblu, float *cen, int *limred, float *direta0, float *dircsn);
	// Ruff initialization fior the strength of the 
	// magnetic field vector
	void RUFFFLD_FC(int *len, float *obsi, float *obsv, int *limblu, 
			float *cen, int *limred, float *cct, float *dlamh, 
			float *disper, float *dircsn, int *ivcen, int *iblu, 
			int *imid, int *ired, float *dirfld);
	void INVERT_FIELD_FC(
			int *nlambda, 
			float *SI, 
			float *SQ, 
			float *SU, 
			float *SV, 
			float *StrayProf, 
			int *continuum1, 
			int *continuum2, 
			int *core1, 
			int *core2, 
			float *Field,
			float *inclination,
			//int * vel, 
			float *vel,
			float *ffactor,
			int *pixel
			);
}



extern "C"
{
	float my_ff(int *n, float* parameters)
	{
		try 
		{

			MERLIN_Init::genetic_model.magnetic_field.value = 500 + (2500. * parameters[0]);
			MERLIN_Init::genetic_model.lines[0].wavelength_shift.value = (2.*(2.*parameters[1]-1.)) * 1.025270223368037;
			MERLIN_Init::genetic_model.lines[1].wavelength_shift.value = MERLIN_Init::genetic_model.lines[0].wavelength_shift.value;
			if(MERLIN_Init::genetic_model.stray_light.free) {
				MERLIN_Init::genetic_model.stray_light.value = parameters[2];
			}
			// Next four lines added back in by SWM to allow the GA to compute the line strength and Doppler Widths
			//MERLIN_Init::genetic_model.doppler_width.value = 10. + (50.*parameters[3]);
			//MERLIN_Init::genetic_model.line_strength.value = 30 * parameters[4];
			//MERLIN_Init::genetic_model.inclination.value = parameters[5] * CSAC::Pi;
			//MERLIN_Init::genetic_model.azimuth.value = parameters[6] * CSAC::Pi;

			//cout<<MERLIN_Init::genetic_model<<endl;
			//exit(1);

			// By RCE: The next 3 statements (lines 1, 2 and 4 from this point) set the stray light
			// model to the ASP formulation (stray light as opposed to filling factor), and store
			// the model set by the user. Then calls the ME synthesis package, computes the profiles
			// and sets the stray light model back to the user defined value. This is because PIKAIA
			// works with the ASP stray light formulation.

			CSAC::stray_light_model_type remember_model = CSAC::stray_light_model;
			CSAC::stray_light_model = CSAC::fraction;
			MERLIN_Init::me.get_profile(MERLIN_Init::genetic_model, MERLIN_Init::trial_profile);
			CSAC::stray_light_model = remember_model;

			double chisq = CSAC::MERLIN_Chisq::compute_chisq(MERLIN_Init::reduced_observed, MERLIN_Init::trial_profile.stokes, MERLIN_Init::genetic_model, 0, MERLIN_Init::continuum_);
			//cout<<"chisq = "<<chisq<<endl;

			return (float) 1/chisq;
		}
		catch(string &error)
		{
			cerr<<"STOP! : Pikaia Callback Exception, this can not be propagated, must exit now."<<endl;
			cerr<<error<<endl;
			exit(1);
		}
		catch (CSAC::MERLIN_Exception &ex)
		{
			cerr<<"STOP! : Pikaia Callback Exception, this can not be propagated, must exit now."<<endl;
			cerr<<ex.get_description()<<endl;
			exit(2);
		}
		catch(...)
		{
			cerr<<"STOP! : Pikaia Callback Exception, this can not be propagated, must exit now."<<endl;
			cerr<<"An undefined exception has ocurred in my_ff call back."<<endl;
			exit(3);
		}
	}
}

namespace MERLIN_Init
{
	CSAC::MERLIN_Model genetic_model; 
	CSAC::MERLIN_MilneEddington me;
	CSAC::MERLIN_StokesProfile reduced_observed;
	CSAC::MERLIN_Profile trial_profile;
	float continuum_;

	void  create_initial_guess(
			float * I,
			float * Q,
			float * U,
			float * V,
			int &len,
			CSAC::MERLIN_StokesProfile &observed,
			float &continuum, 
			CSAC::MERLIN_Model &my_model,
			double &genetic_chisq,
			double &netpol,
			int &individuals,
			int &generations
			)
	{

		// Required by the ruff initialization
		float line_center = 75.78;
		int line_center_appx_pixel = 76;
		int limblu = 61;
		int limred = 91;

		// DLSP Dispersion
		// float dispersion = 21.562;
		// HINODE Dispersion
		const float dispersion =21.549;
		const float dvdmin = 20.0;

		float obsdp;
		float sb1mu;
		float sbzero;
		float azm;

		// Parameter space
		int n=3;
		float *parameters;
		parameters = new float [n];

		float ctrl [12];

		float fit;

		float min_doppler_width = dvdmin/dispersion;

		// status of pikaia
		int status;

		// In fortran number style for arrays
		// ctrl(1) - number of individuals in a population (default is 100)
		// ctrl(2) - number of generations over which solution is to evolve (default is 500)

		ctrl[0]  = individuals;
		ctrl[1]  = generations;
		// Set all other values to -1
		for (int ll=2; ll < 12; ll++)
			ctrl[ll] = -1;
		// SMC: Added Creep Mutation
		ctrl[4] = 5;
		// SMC: Run in "Elitist" Mode
		ctrl[10] = 1;

		genetic_model = my_model;
		reduced_observed = observed;
		continuum_ = continuum;
		//MERLIN_Init::reduced_observed.print_stokes_vectors(cout,"s_lambda","s_I","s_Q","s_U","s_V");

		if (genetic_model.lines.size()  != 2)
			throw string("There are more atomic lines than the allowed for initialization in this code.");

		// The macroturbulence does not really matter
		// set it to zero so it speed up the genetic part
		genetic_model.vmac.value =    0.0;

		// Set the elements of the source function
		obsdp = continuum - I[line_center_appx_pixel];
		sb1mu = obsdp < (0.8 * 0.9999 * continuum) ? obsdp : (0.8 * 0.9999 *continuum);
		sbzero = continuum - sb1mu;
		genetic_model.planck_function_gradient.value  = sb1mu;
		genetic_model.planck_function.value = 0.6 * sbzero;
		genetic_model.damping.value = 0.3;
		genetic_model.doppler_width.value = 25;

		float direta0;
		float dircsn;
		RUFFCSN_FC(&len, &Q[0], &U[0], &V[0], &limblu, &line_center, &limred, &direta0, &dircsn);
		genetic_model.line_strength.value  = direta0;
		genetic_model.inclination.value = acos(dircsn);

		//Call the ruff guess for the azimuth
		//cout<<"len = "<<len<<endl;
		RUFFAZM_FC(&len, &Q[0], &U[0], &limblu, &line_center, &limred, &min_doppler_width, &azm);
		if (azm < 0)
			genetic_model.azimuth.value = azm + CSAC::Pi;
		else
			genetic_model.azimuth.value = azm;
		//cout<<"azm = "<<genetic_model.azimuth.value*(180/CSAC::Pi)<<endl;

		// Call the neural network
		// Neural Network Init values
		// float field;
		//         float vel; 
		//         float Ffactor;
		//         float inclination;
		//         int continuum1 = 54;
		//         int continuum2 = 64; 
		//         int core1 = 19;
		//         int core2 = 41;  
		//         float *slgt = new float[len];  
		//         for(int u=0; u < len; u++)
		//             slgt[u] = (float) my_model.stray_light_profile[u];
		//         int iii = 1;
		//         INVERT_FIELD_FC(&len, I, Q, U ,V, slgt, &continuum1, &continuum2, &core1, &core2, &field, &inclination, &vel, &Ffactor, &iii);
		//         genetic_model.inclination.value = (CSAC::Pi/180)*inclination;

		//genetic_model.stray_light.value = 0.0; //1-Ffactor;

		genetic_model.stray_light_shift.value = 0.0;

		// By RCE: We comment out the statement that generates a random seed for PIKAIA.
		// We set initial seed to 12345 for testing. 
		// Initialize the random number generator
		// int seed = getpid();
		int seed = 12345;
		srand(seed);

		// Initialize the random number generator for pikaia
		seed = abs((int)rand());
		if (seed <= 0)
			throw string("Invalid seed for random number generator");
		RNINIT_FC(&seed);

		// Call pikaia.
		PIKAIA_FC(my_ff , &n, ctrl, parameters, &fit, &status);

		if (status == 0)
		{
			// Recover the best individual returned by Pikaia
			genetic_model.magnetic_field.value = 500 + (2500. * parameters[0]);
			cout<<"+"<<flush;
			//cout<<"Fit = "<<1/fit<<endl;
			//exit(1);
			if(genetic_model.magnetic_field.value > 3000)
			{
				string error ("Internal Error: The returned value genetic init for the magnetic field is larger than 3000 Gauss. STOP!");
				throw error;
			}
			if ( (netpol <0.06) && (continuum < 3000) && (genetic_model.magnetic_field.value < 2000))
				genetic_model.magnetic_field.value = 2800;

			genetic_model.lines[0].wavelength_shift.value = (2.*(2.*parameters[1]-1.)) * 1.025270223368037;
			genetic_model.lines[1].wavelength_shift.value = genetic_model.lines[0].wavelength_shift.value;
			// Next four lines added back in by SWM to allow the GA to compute the line strength and Doppler Widths
			//genetic_model.doppler_width.value =  10. + (50.*parameters[3]);
			//genetic_model.line_strength.value = 30 * parameters[4];
			//genetic_model.inclination.value = parameters[5] * CSAC::Pi;
			//genetic_model.azimuth.value = parameters[6] * CSAC::Pi;

			if (genetic_model.stray_light.free) {
				if (CSAC::stray_light_model == CSAC::fill)
				{
					genetic_model.stray_light.value = 1 - parameters[2];

					// The next two instructions were added by RCE to correct for the stray light factor in
					// the Planck function
					genetic_model.planck_function_gradient.value = genetic_model.planck_function_gradient.value / (1 - parameters[2]);
					genetic_model.planck_function.value = genetic_model.planck_function.value / (1 - parameters[2]);
				}
				else if (CSAC::stray_light_model == CSAC::fraction)
					genetic_model.stray_light.value = parameters[2];
				else
				{
					char error_line[32];
					sprintf(error_line, "%i", __LINE__);
					string t_error=__FILE__ + string(":") + string(error_line) + ": the stray_light_model is invalid\n";
					throw t_error;
				}
			}
			else
			{
				// even if the stray light is not free, make sure we set the parameter correctly
				if (CSAC::stray_light_model == CSAC::fill)
				{
					genetic_model.stray_light.value = 1;
				}
				else if (CSAC::stray_light_model == CSAC::fraction)
					genetic_model.stray_light.value = 0;
				else
				{
					char error_line[32];
					sprintf(error_line, "%i", __LINE__);
					string t_error=__FILE__ + string(":") + string(error_line) + ": the stray_light_model is invalid\n";
					throw t_error;
				}
			}

			// vmac for DLSP
			//genetic_model.vmac.value =   1.114;
			// vmac for HINODE
			genetic_model.vmac.value =    0.7083;

			// Get the initialization fit
			me.get_profile(genetic_model, trial_profile);
			genetic_chisq = CSAC::MERLIN_Chisq::compute_chisq(reduced_observed, trial_profile.stokes, genetic_model, 0, continuum);
			my_model = genetic_model;
		}
		else
		{
			string error("Internal Error: Pikaia is not happy, STOP!");
			throw error;
		}

		delete [] parameters;
	}
}




