/*
 *  Language: C++
 *  File: extcon.cc
 *  Package: Milne Eddington gRid Linear Inversion Network (MERLIN) 
 *
 *  Created by Jose Garcia on 11/11/06.
 *  Copyright 2006 UCAR. All rights reserved.
 *
 */

#include <math.h>

#include <iostream>
#include <fstream>
#include <vector>
#include "CSACFile.h"

#include <netcdf.h>

using namespace std;

bool merlin_debug_ = false;

#define MINIMUM_POLARIZATION 0.006

const double extcon_pi = atan((double)1.0)*4.0;  

double get_net_polarization(vector<double> &I, vector<double> &Q, vector<double> &U, vector<double> &V)
{

    // Determine the max value of I for the observed continuum
    double Imax = I[0];
    for (int y=1; y < I.size(); y++)
    {
        if (Imax < I[y])
            Imax = I[y];
    }
  

    double I_sum = 0.0;
    double Q_sum = 0.0;
    double U_sum = 0.0;
    double V_sum = 0.0;
	
	
    int n_data = 0;
	
    for (int i=0; i < I.size(); i++) 
    {
		
        I_sum += I[i];   
        Q_sum += Q[i] * Q[i];    
        U_sum += U[i] * U[i];     
        V_sum += V[i] * V[i];
			
        n_data++;

    }

 
    // normalize data
    I_sum = I_sum / n_data;
    Q_sum = Q_sum / n_data;
    U_sum = U_sum / n_data;
    V_sum = V_sum / n_data;

    //cout<<"NET_POL "<<( sqrt(Q_sum + U_sum + V_sum) / I_sum )<<" for ndata "<<n_data<<endl;
	
    return ( sqrt(Q_sum + U_sum + V_sum) / I_sum );
}

inline double get_max_I(const vector <double> &vec)
{
    int my_size = vec.size();
    if (my_size < 1)
        throw string("The provided vector is wrong, min size must be greater than one");
	
    double max = vec[0];
    for (int i=0; i < my_size; i++)
    {
        if (max < vec[i])
            max = vec[i];
    }
    return max;
}

inline double get_avg_V(const vector<double> &vec)
{
  
    int my_size = vec.size();
    if (my_size < 1)
        throw string("The provided vector is wrong, min size must be greater than one"); 
  
    double sum =0.0;
    for (int i=0; i < my_size; i++)
        sum+=vec[i];
    return sum;// / my_size;
}


void compute_smoooth_extension (vector<double> &stray_light, vector <double> &extension)
{
    int power_of_two = 4;
    while (power_of_two < stray_light.size())
        power_of_two *=2;
    //cout<<"PO2: "<<power_of_two<<endl;
    int extension_size = power_of_two - stray_light.size();
    double arg;
    extension.reserve(extension_size);
    for (int i=0; i < extension_size; i++)
    {
        //arg = .5 * (1 -  cos((1:extension)*pi/extension))
        arg = .5 * (1 - cos((i+1) * extcon_pi / extension_size));
        //signal(columns+1:my_pow) = arg * signal(1) + (1-arg) * signal(columns);
        extension.push_back( arg * stray_light[0] + (1-arg) * stray_light[stray_light.size() -1]);
    }
    
}


void build_continuum_and_stray_light(vector<string> &map, int x1, int x2, int y1, int y2)
{
    // prepare the continuum matrix 
    vector < vector <double> > continuum;
    // Data structures for other stuff
    vector < vector <double> > net_pol;
    vector < vector <double> > stokes_V;
    // Get data structure ready for work
    vector <double> dummy;
    continuum.resize(map.size(),dummy);
    stokes_V.resize(map.size(),dummy);
    net_pol.resize(map.size(),dummy);
    // Data structure used to store
    // the position of each Stokes I
    // vector chosen in the calculation
    // of the Stray Light profile.
    vector< vector<int> > chosen;
    chosen.resize( map.size());

    double net_pol_value;
  
    // Prepare for the stray light
    vector<double> stray_light;

    int counter = 0;
  
    CSACFile solarb;

    for (int i=0; i < map.size(); i++)
    {
        cout << "Procesing " << map[i] << endl;
        solarb.open(map[i]);
        cout<<"Size of the number of slit positions dimension "<<solarb.get_number_of_pixels()<<endl;
        cout<<"Size of I, Q, U and V is "<<solarb.get_size_of_vectors()<<endl;
    
        // we resize the stray light vector only once
        if (i == 0)
            stray_light.resize(solarb.get_size_of_vectors(),0.0);
    
        continuum[i].resize(solarb.get_number_of_pixels(),0.0);
        stokes_V[i].resize(solarb.get_number_of_pixels(),0.0);
        net_pol[i].resize(solarb.get_number_of_pixels(),0.0);
        chosen[i].resize(solarb.get_number_of_pixels(), 0);

        //cout<<"The current size of continuum[i] is "<<continuum[i].size()<<endl;
        vector<double> current_I;
        vector<double> current_Q;
        vector<double> current_U;
        vector<double> current_V;

        //current_I.resize(solarb.get_size_of_vectors(),0.0);
        // Loop over all slit positions
        
        //cout<<"NET_POL as saved in vector is "<< net_pol[i][j]<<endl;


        for (int j=0; j < solarb.get_number_of_pixels(); j++)
        {
            solarb.get_I(j+1, current_I);
            solarb.get_Q(j+1, current_Q);
            solarb.get_U(j+1, current_U);
            solarb.get_V(j+1, current_V);

            net_pol_value = get_net_polarization( current_I, current_Q, current_U, current_V);
            net_pol[i][j]=net_pol_value;


            // Are we in the region for extracting the stray light?
            if (i+1 >= x1 && i+1 <= x2) 
            {
                if (j+1 >=y1 && j+1 <= y2)
                {
                    // cout<<"File selected for stray light determination"<<endl;
                    // Yes, we are, we add this continuum to the average stray light vector
                    // we get sure that the current I is the same size as the stray light as 
                    // it was set for the first file in the map
                    if (stray_light.size() != current_I.size())
                    {
                        cerr<<"STOP! Size missmatch detected"<<endl;
                        cerr<<"The size of the stray light vector is "<<stray_light.size()<<endl;
                        cerr<<"The size of the current I is "<<current_I.size()<<endl;
                        exit(1);
                    }
                    // Everything OK, add ...
                    if (  net_pol_value < MINIMUM_POLARIZATION)
                    {
                        chosen[i][j]=1;
                        //cout<<"In X = "<<i<<" and Y = "<<j<<endl;
                        //cout<<"Got one with value "<<net_pol_value<<endl;
                        for (int p=0; p < current_I.size(); p++) 
                            stray_light[p] += current_I[p]; 
                        counter++;
                    }
                }
            }
            // Add the maximum of current I to the continuum
            continuum[i][j] = get_max_I(current_I);
            // Add the average V to the V image;
            stokes_V[i][j] = get_avg_V(current_V);
        }
    }
  
    // Divide the stray light by the counter to obtain the average
    if (counter > 1)
    {
        for (int p=0; p < stray_light.size(); p++) 
        {
            stray_light[p] =   stray_light[p] / counter;
            // print the stray light vector to standard output
            cout<<stray_light[p]<<' ';
        }
    }
    cout<<endl;

    vector <double> extension;
    compute_smoooth_extension(stray_light, extension);

    cout<<"SL extension: "<<endl;
    for (int t=0; t < extension.size(); t++)
        cout<<extension[t]<<" "<<flush;
    cout<<endl;

  

    int ncid;
    nc_create("images.nc",NC_CLOBBER,&ncid);
    int dim_ids[2];
    int status;
    status = nc_def_dim(ncid,"X",map.size(), &dim_ids[0]);
    status = nc_def_dim(ncid,"Y",solarb.get_number_of_pixels(), &dim_ids[1]);
    int lambda_id;
    status = nc_def_dim(ncid,"Lambda",solarb.get_size_of_vectors(), &lambda_id);
    int extension_id;
    status = nc_def_dim(ncid,"extension_size",extension.size(), &extension_id);
    int cont_id;
    status = nc_def_var(ncid, "continuum", NC_DOUBLE, 2, dim_ids, &cont_id);
    int stokes_V_id;
    status = nc_def_var(ncid, "stokes_V", NC_DOUBLE, 2, dim_ids, &stokes_V_id);
    int net_pol_id;
    status = nc_def_var(ncid, "net_polarization", NC_DOUBLE, 2, dim_ids, &net_pol_id);
    int sl_id;
    status = nc_def_var(ncid, "stray_light", NC_DOUBLE, 1, &lambda_id, &sl_id);
    int sl_ext_id;
    status = nc_def_var(ncid, "stray_light_extension", NC_DOUBLE, 1, &extension_id, &sl_ext_id);
    int chosen_id;
    status = nc_def_var(ncid, "chosen", NC_INT, 2, dim_ids, &chosen_id);

    nc_enddef(ncid);

    size_t start[2];
    start[0]=0;
    start[1]=0;
    size_t count[2];
    count[0] = 1;
    count[1] = solarb.get_number_of_pixels();
    for(int g=0; g < map.size(); g++)
    {
        start[0]= g;
        status = nc_put_vara_double(ncid, cont_id, start, count, &continuum[g][0]);
        status = nc_put_vara_double(ncid, stokes_V_id, start, count, &stokes_V[g][0]);
        status = nc_put_vara_double(ncid, net_pol_id, start, count, &net_pol[g][0]);
        status = nc_put_vara_int(ncid, chosen_id, start, count, &chosen[g][0]);
    }
    nc_put_var_double(ncid, sl_id, &stray_light[0]);
    nc_put_var_double(ncid, sl_ext_id, &extension[0]);
    nc_close(ncid);
}

int main(int argc, char* argv[])
{
    if (argc != 6)
    {
        cerr << argv[0] << ": [map_file] [X1] [X2] [Y1] [Y2]"<<endl;
        return 1;
    }
    ifstream map_file;
    map_file.open(argv[1]);
    if (!map_file)
    {
        cerr << argv[0] <<": The map file " << argv[1] << " does not exist or can not be open." << endl;
        return 1;
    }
    string my_file;
    vector<string> the_map;
    while(map_file)
    {
        getline(map_file, my_file);
        if (my_file != "")
            the_map.push_back(my_file);
    }
    map_file.close();
     cout<<"Notice, the stray light vector will be build from\n"
        <<"X1= "<<atoi(argv[2])<<" X2 = "<<atoi(argv[3])<<endl
        <<"Y1= "<<atoi(argv[4])<<" Y2 = "<<atoi(argv[5])<<endl
        <<"countinue? (Y/y)?"<<endl;
      string answer;
      cin>>answer;
      if(answer != "Y" && answer != "y")
      {
        cout<<"You have selected "<<answer<<endl;
        return 0;
      }
    try
    {
        build_continuum_and_stray_light(the_map,atoi(argv[2]),atoi(argv[3]),atoi(argv[4]),atoi(argv[5]));
    }
    catch(string &ex)
    {
        cerr<<"Exception :"<<ex<<endl;
    }
    catch(...)
    {
        cerr<<"Unknown exception"<<endl;
    }
}



