/* ============================================================================
 * Language: C++ 
 * File: MERLIN.h
 * Package Name: Milne Eddington gRid Linear Inversion Network (MERLIN).
 * Sub-package: Fundamental Inversion Code (FICO).
 *
 * Created by Jose Garcia <jgarcia@ucar.edu> on 11/11/05.
 *  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact University Corporation for Atmospheric Research at
 * 3080 Center Green Drive, Boulder, CO 80301

 * (c) COPYRIGHT University Corporation for Atmospheric Research 2004-2006
 * Please read the full copyright statement in the file COPYRIGHT_UCAR.
 *
 =========================================================================== */



/** \mainpage Index Page: Milne Eddington Inversions with MERLIN
 *
 * \section intro_sec Introduction
 *
 * MERLIN is a framework developed to allow scientist to invert
 * observed spectral lines  based on a theoretical Milne Eddington
 * model and inverted using the Levenberg-Marquardt algorithm.
 * 
 *
 * \section install_sec Installation
 * Refer to the document INSTALL provided with the distribution.
 *
 * \subsection step1 Step 1: Coding with MERLIN
 * Please write to Jose Garcia (jgarcia@ucar.edu)
 * To get some examples of applications coded using the MERLIN
 * framework, for example, how to use the Voigt function stand alone,
 * how to compute the Zeeman effect, just synthesizing Stokes
 * profiles for a given ME atmosphere, etc.
 *  
 */

#ifndef MERLIN_H_
#define MERLIN_H_ 1

#include "config.h"

#include <math.h>
#include "errno.h"

#ifndef SIZEOF_UNSIGNED_CHAR
#error Which is the size for the MERLIN datatype "byte" in this machine?
#else 
#define MERLIN_HAVE_SIZE_OF_UNSIGNED_CHAR
#endif

#ifndef SIZEOF_SHORT_INT 
#error Which is the size for the MERLIN datatype "field" in this machine?
#else 
#define MERLIN_HAVE_SIZE_OF_SHORT_INT
#endif

#if defined(MERLIN_HAVE_SIZE_OF_UNSIGNED_CHAR) && defined(MERLIN_HAVE_SIZE_OF_SHORT_INT)

#if REQUIRED_SIZEOF_UNSIGNED_CHAR != SIZEOF_UNSIGNED_CHAR
#error The calculated size for the MERLIN datatype "byte" in this machine is not supported!
#endif


#if REQUIRED_SIZEOF_SHORT_INT != SIZEOF_SHORT_INT
#error The calculated size for the MERLIN datatype "field" in this machine is not supported!
#endif

#endif


#ifndef MERLIN_SINGLE_PRECISION
typedef double merlin_real;
#else
typedef float  merlin_real;
#endif // MERLIN_SINGLE_PRECISION


namespace CSAC
{
    /// controls the maximun damping parameter for the Levenberg-Marquardt algorithm.
    extern double max_damping_parameter;
	
    /// controls the minimum damping parameter for the Levenberg-Marquardt algorithm.
    extern double min_damping_parameter;
	
    /// maximun number of allowed failures before the inversion algorithm kicks out.
    extern int max_number_failures;
	
    /// maximum number of iterations in the Levenberg-Marquardt algorithm.
    extern int max_number_iterations;

    /// This is a critical variable in the inversion
    /// as it controls when we reject the matrix A
    /// in the linear system Ax = b. The matrix A
    /// is the Jacobian of the current fit.
    /// See the Levenberg-Marquardt algorithm. 
    extern double worst_tolerated_condition_number;

    // You have no idea how important this variable really is !
    // It controls the path of descend to the inversion method
    // and it is highly realted to the data
    // extern double SVD_threshold;

    /// Control portions of code designed to track down bugs
    extern bool merlin_debug;

    /// For fill factor 
    /// Stokes_I = alpha * Stokes_I + (1 - alpha) * Stray_light 
#ifndef STRAY_LIGHT_MODEL_FILL_FACTOR
#define STRAY_LIGHT_MODEL_FILL_FACTOR 0
#endif
    /// For fraction
    /// Stokes_I = Stokes_I + alpha * Stray_light
#ifndef STRAY_LIGHT_MODEL_FRACTION
#define STRAY_LIGHT_MODEL_FRACTION 1 
#endif

    /// Only to values should be assigned to the stray_light_model, 0 or 1
    enum  stray_light_model_type {  fill = STRAY_LIGHT_MODEL_FILL_FACTOR, fraction = STRAY_LIGHT_MODEL_FRACTION};

    /// Controls the model used for adding the stray light
    /// to the theoretical profiles
    extern stray_light_model_type stray_light_model;
    
    /// Weights to control the inversion
    
    extern double STOKES_I_WEIGHT; 
    
    extern double STOKES_Q_WEIGHT;
    
    extern double STOKES_U_WEIGHT;
    
    extern double STOKES_V_WEIGHT;   

    /// Speed of light (m/s) 
    extern const merlin_real CC; 
    /// Value of \f$ \pi \f$ 
    extern const merlin_real Pi;  
    /// \brief Used in the computation of the Faraday function.
    /// @see mvoigt.cc
    extern const merlin_real Piis; 
  
    /// 
    extern const merlin_real perturbation;
  
	
    /// If computing a numerica derivative, this function
    /// is used to determine the order of magnitude of the
    /// value f(x) in order to obtain a sensitive 
    /// perturbation h to produce the computation
    /// (f(x+h) - f(x)) / h
    inline int order_of_magnitude (double val)
    {
        // Note: this is not technically correct but for the purpose 
        // of this function, it works just fine
        if (val == 0 )
            return 0;
		
        val = fabs(val);
        int order=0;
        if (val == 1)
            return order;
        else if (val < 1)
        {
            while (val < 1)
            {
                val = val * 10;
                order --;
            }
            return order;
        }
        else // val must be greater than one
        {
            while (val >= 10)
            {
                val = val / 10;
                order ++;
            }
            return order;
        }
    }

	
	
}; // end namespace CSAC

/// Substract the value of time timeval structure for resource usage.
extern "C"
{
  int
  timeval_subtract (struct timeval *result, struct timeval*x, struct timeval*y);
}

#endif // MERLIN_H_ 
