/* ============================================================================
* Language: C++ 
* File: MERLIN_Chisq.h
* Package Name: Milne Eddington gRid Linear Inversion Network (MERLIN).
* Sub-package: Fundamental Inversion Code (FICO).
*
* Created by Jose Garcia <jgarcia@ucar.edu> on 11/11/05.
*  
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
* 
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
* 
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* You can contact University Corporation for Atmospheric Research at
* 3080 Center Green Drive, Boulder, CO 80301

* (c) COPYRIGHT University Corporation for Atmostpheric Research 2004-2006
* Please read the full copyright statement in the file COPYRIGHT_UCAR.
*
=========================================================================== */

#include "MERLIN.h"
#include "MERLIN_Chisq.h"

using CSAC::MERLIN_Chisq;
using CSAC::MERLIN_StokesProfile;
using CSAC::MERLIN_Model;


// SAG: Compute CHI^2. Returns Chi^2 and the scale factor 
MERLIN_Chisq
MERLIN_Chisq::compute_chisq(
                            MERLIN_StokesProfile &observed, 
                            MERLIN_StokesProfile &synthetic, 
                            MERLIN_Model &guess, 
                            MERLIN_StokesProfile *sigma,
                            const double &normalization)
{
    
    double chisq = 0.0;
    double diff=0.0;
    unsigned int n_data=0;
    if(sigma)
    {
        
        if (guess.vectors_to_invert.invert_I)
        {
            for (int i=0; i < guess.n_total_wave; i++)
            {
                diff=(observed.I[i] - synthetic.I[i]) * STOKES_I_WEIGHT;
                chisq+=(diff * diff)/(sigma->I[i]*sigma->I[i]);
            }
            n_data+=guess.n_total_wave;
        }
        if (guess.vectors_to_invert.invert_Q)
        {
            for (int i=0; i < guess.n_total_wave; i++)
            {
                diff=(observed.Q[i] - synthetic.Q[i]) * STOKES_Q_WEIGHT;
                chisq+=(diff * diff)/(sigma->Q[i]*sigma->Q[i]);
            }
            n_data+=guess.n_total_wave;
        }
        
        if (guess.vectors_to_invert.invert_U)
        {
            for (int i=0; i < guess.n_total_wave; i++)
            {
                diff=(observed.U[i] - synthetic.U[i]) * STOKES_U_WEIGHT;
                chisq+=(diff * diff)/(sigma->U[i]*sigma->U[i]);
            }
            n_data+=guess.n_total_wave;
        }
        if (guess.vectors_to_invert.invert_V)
        {
            for (int i=0; i < guess.n_total_wave; i++)
            {
                diff=(observed.V[i] - synthetic.V[i]) * STOKES_V_WEIGHT;
                chisq+=(diff * diff)/(sigma->V[i]*sigma->V[i]);
            }
            n_data+=guess.n_total_wave;
        }
    }
    else // no sigma was provided
    {
        if (guess.vectors_to_invert.invert_I && guess.vectors_to_invert.invert_Q && guess.vectors_to_invert.invert_U && guess.vectors_to_invert.invert_V)
        {
            for (int i=0; i < guess.n_total_wave; i++)
            {
                diff = (observed.I[i] - synthetic.I[i]);
                chisq += (diff * diff) * STOKES_I_WEIGHT;
                diff = (observed.Q[i] - synthetic.Q[i]);
                chisq += (diff * diff) * STOKES_Q_WEIGHT;
                diff = (observed.U[i] - synthetic.U[i]);
                chisq += (diff * diff) * STOKES_U_WEIGHT;
                diff = (observed.V[i] - synthetic.V[i]);
                chisq += (diff * diff) * STOKES_V_WEIGHT;
            }
            n_data+=guess.n_total_wave * 4;
            
        }
        else
        {
            
            if (guess.vectors_to_invert.invert_I)
            {
                for (int i=0; i < guess.n_total_wave; i++)
                {
                    diff=(observed.I[i] - synthetic.I[i]);
                    chisq+=(diff * diff) * STOKES_I_WEIGHT;
                }
                n_data+=guess.n_total_wave;
            }
            if (guess.vectors_to_invert.invert_Q)
            {
                for (int i=0; i < guess.n_total_wave; i++)
                {
                    diff=(observed.Q[i] - synthetic.Q[i]);
                    chisq+=(diff * diff) * STOKES_Q_WEIGHT;
                }
                n_data+=guess.n_total_wave;
            }
            if (guess.vectors_to_invert.invert_U)
            {
                for (int i=0; i < guess.n_total_wave; i++)
                {
                    diff=(observed.U[i] - synthetic.U[i]);
                    chisq+=(diff * diff) * STOKES_U_WEIGHT;
                }
                n_data+=guess.n_total_wave;
            }
            if (guess.vectors_to_invert.invert_V)
            {
                for (int i=0; i < guess.n_total_wave; i++)
                {
                    diff=(observed.V[i] - synthetic.V[i]);
                    chisq+=(diff * diff) * STOKES_V_WEIGHT;
                }
                n_data+=guess.n_total_wave;
            }
        }
    }
    // renormalize according to the weighted average
    chisq = chisq / (STOKES_I_WEIGHT + STOKES_Q_WEIGHT + STOKES_U_WEIGHT + STOKES_V_WEIGHT);
    // renormalize according to the number of samples in the vectors I, Q, U and V
    chisq = chisq / (n_data * normalization * normalization);
    
    // SAG: Recompute scale factor
    double scale_factor;
    scale_factor  = (STOKES_I_WEIGHT + STOKES_Q_WEIGHT + STOKES_U_WEIGHT + STOKES_V_WEIGHT);
    scale_factor *= n_data;
    scale_factor *= normalization * normalization;
    
    // SAG: Output
    MERLIN_Chisq output;
    output.set_chisq(chisq);
    output.set_scale_factor(scale_factor);
    
    return output;
}

double 
MERLIN_Chisq::compute_chisq_I(
                            MERLIN_StokesProfile &observed, 
                            MERLIN_StokesProfile &synthetic, 
                            MERLIN_Model &guess, 
                            MERLIN_StokesProfile *sigma,
                            const double &normalization)
{
    double chisq = 0.0;
    double diff=0.0;
    unsigned int n_data=0;
    
    // If Stokes I is set not to be inverted, then
    // we return chisq = -1 as to indicate
    // that there is not possible value for this
    // request.
    if (!guess.vectors_to_invert.invert_I)
        return -1;
    
    if(sigma)
    {
        for (int i=0; i < guess.n_total_wave; i++)
        {
            diff=(observed.I[i] - synthetic.I[i]) * STOKES_I_WEIGHT;
            chisq+=(diff * diff)/(sigma->I[i]*sigma->I[i]);
        }
        n_data+=guess.n_total_wave;
    }
    else // no sigma was provided
    {
            for (int i=0; i < guess.n_total_wave; i++)
            {
                diff = (observed.I[i] - synthetic.I[i]);
                chisq += (diff * diff) * STOKES_I_WEIGHT;
            }
            n_data+=guess.n_total_wave ;
    }
    // renormalize according to the weighted average
    chisq = chisq / (STOKES_I_WEIGHT);
    // renormalize according to the number of samples in the vector I
    chisq = chisq / (n_data * normalization * normalization);
    
    return chisq;
}

double 
MERLIN_Chisq::compute_chisq_Q(
                            MERLIN_StokesProfile &observed, 
                            MERLIN_StokesProfile &synthetic, 
                            MERLIN_Model &guess, 
                            MERLIN_StokesProfile *sigma,
                            const double &normalization)
{
    double chisq = 0.0;
    double diff=0.0;
    unsigned int n_data=0;
    
    // If Stokes Q is set not to be inverted, then
    // we return chisq = -1 as to indicate
    // that there is not possible value for this
    // request.
    if (!guess.vectors_to_invert.invert_Q)
        return -1;
    
    if(sigma)
    {
        for (int i=0; i < guess.n_total_wave; i++)
        {
            diff=(observed.Q[i] - synthetic.Q[i]) * STOKES_Q_WEIGHT;
            chisq+=(diff * diff)/(sigma->Q[i]*sigma->Q[i]);
        }
        n_data+=guess.n_total_wave;
    }
    else // no sigma was provided
    {
            for (int i=0; i < guess.n_total_wave; i++)
            {
                diff = (observed.Q[i] - synthetic.Q[i]);
                chisq += (diff * diff) * STOKES_Q_WEIGHT;
            }
            n_data+=guess.n_total_wave ;
    }
    // renormalize according to the weighted average
    chisq = chisq / (STOKES_Q_WEIGHT);
    // renormalize according to the number of samples in the vector Q
    chisq = chisq / (n_data * normalization * normalization);
    
    return chisq;
}

double 
MERLIN_Chisq::compute_chisq_U(
                            MERLIN_StokesProfile &observed, 
                            MERLIN_StokesProfile &synthetic, 
                            MERLIN_Model &guess, 
                            MERLIN_StokesProfile *sigma,
                            const double &normalization)
{
    double chisq = 0.0;
    double diff=0.0;
    unsigned int n_data=0;
    
    // If Stokes U is set not to be inverted, then
    // we return chisq = -1 as to indicate
    // that there is not possible value for this
    // request.
    if (!guess.vectors_to_invert.invert_U)
        return -1;
    
    if(sigma)
    {
        for (int i=0; i < guess.n_total_wave; i++)
        {
            diff=(observed.U[i] - synthetic.U[i]) * STOKES_U_WEIGHT;
            chisq+=(diff * diff)/(sigma->U[i]*sigma->U[i]);
        }
        n_data+=guess.n_total_wave;
    }
    else // no sigma was provided
    {
            for (int i=0; i < guess.n_total_wave; i++)
            {
                diff = (observed.U[i] - synthetic.U[i]);
                chisq += (diff * diff) * STOKES_U_WEIGHT;
            }
            n_data+=guess.n_total_wave ;
    }
    // renormalize according to the weighted average
    chisq = chisq / (STOKES_U_WEIGHT);
    // renormalize according to the number of samples in the vectors U
    chisq = chisq /  (n_data * normalization * normalization);
    
    return chisq;
}

double 
MERLIN_Chisq::compute_chisq_V(
                            MERLIN_StokesProfile &observed, 
                            MERLIN_StokesProfile &synthetic, 
                            MERLIN_Model &guess, 
                            MERLIN_StokesProfile *sigma,
                            const double &normalization)
{
    double chisq = 0.0;
    double diff=0.0;
    unsigned int n_data=0;
    
    // If Stokes V is set not to be inverted, then
    // we return chisq = -1 as to indicate
    // that there is not possible value for this
    // request.
    if (!guess.vectors_to_invert.invert_V)
        return -1;
    
    if(sigma)
    {
        for (int i=0; i < guess.n_total_wave; i++)
        {
            diff=(observed.V[i] - synthetic.V[i]) * STOKES_V_WEIGHT;
            chisq+=(diff * diff)/(sigma->V[i]*sigma->V[i]);
        }
        n_data+=guess.n_total_wave;
    }
    else // no sigma was provided
    {
            for (int i=0; i < guess.n_total_wave; i++)
            {
                diff = (observed.V[i] - synthetic.V[i]);
                chisq += (diff * diff) * STOKES_V_WEIGHT;
            }
            n_data+=guess.n_total_wave ;
    }
    // renormalize according to the weighted average
    chisq = chisq / (STOKES_V_WEIGHT);
    // renormalize according to the number of samples in the vectors V
    chisq = chisq /  (n_data * normalization * normalization);
    
    return chisq;
}
