/* ============================================================================
 * Language: C++ 
 * File: MERLIN_InversionStepWiseRegression.h
 * Package Name: Milne Eddington gRid Linear Inversion Network (MERLIN).
 * Sub-package: Fundamental Inversion Code (FICO).
 *
 * Created by Jose Garcia <jgarcia@ucar.edu> on 05/15/07.
 *  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact University Corporation for Atmospheric Research at
 * 3080 Center Green Drive, Boulder, CO 80301

 * (c) COPYRIGHT University Corporation for Atmostpheric Research 2004-2006
 * Please read the full copyright statement in the file COPYRIGHT_UCAR.
 *
 =========================================================================== */

#ifndef MERLIN_InversionStepWiseRegression_h_
#define MERLIN_InversionStepWiseRegression_h_ 1

#include <math.h>

#include "MERLIN.h"
#include "MERLIN_MilneEddington.h"
#include "MERLIN_Profile.h"
#include "MERLIN_ModelCorrection.h"
#include "MERLIN_Inversion.h"

namespace CSAC
{
    
    ///
    class MERLIN_InversionStepWiseRegression: public MERLIN_Inversion
    {
    private:
        ///
        double _continuum_normalization;
        /// maximun number of iterations allowed, set by default to 70
        int _max_iters;
        /// Why did the inversion stop?
        exit_condition _why_did_we_stop;
        /// minimum damping parameter
        double _min_damping_parameter;
        /// maximum damping parameter
        double _max_damping_parameter;
        /// maximum number of consecutive failures 
        int _max_number_failures;
    public:
        ///
        MERLIN_InversionStepWiseRegression(){
            
        };
        ///
        void set_continuum_normalization(const double &v) { _continuum_normalization = v; }
        ///
        void zero_dependent_variables(double ** A, int * used, const int &dimension)
        {
            for (int k = 0; k < dimension; k++)
            {
                if (!used[k])
                    A[k][k] = 0.0 < A[k][k] ? A[k][k] : 0.0;
                else
                    A[k][k] = 0.0 < A[k][k] ? 0.0 : A[k][k];
            }
        }
        ///
        double fprob(const double &f, const double &ss1, const double &ss2)
        {
            if ( (ss1 <= 0.0) || (ss2 <= 0.0) || (f <= 0.0) )
                return 1.0;
            double g = pow(f,.33333333);
            double s1 = .22222222 / ss1;
            double s2 = .22222222 / ss2;
            double u = ( (1.0 - s2) * g - 1.0 + s1) / sqrt(2.0 * (s2 * pow(g,2) + s1));
            return fabs(0.5 * (1.0 - erf(u)));
        }
        /// SAG: Compute parameter uncertainty 
        void compute_uncertainty(
                                 MERLIN_Model &guess,
                                 MERLIN_Model &trial,
                                 MERLIN_StokesProfile &observed_profile,
                                 MERLIN_Profile &trial_profile, 
                                 int total_matrix_size, 
                                 double chisq,
                                 double scale_factor);        
        ///
        virtual double execute (MERLIN_StokesProfile &observed_profile, 
                        MERLIN_Model &guess, 
                        MERLIN_StokesProfile *sigma, 
                        int &n_iterations,
                        double &chisq_I,
                        double &chisq_U,
                        double &chisq_Q,
                        double &chisq_V);
        /// change the default number of maximun iterations
        virtual void set_maximun_iterations (const int &max) { _max_iters = max; }
        /// return the reason why the inversion stopped
        virtual exit_condition get_exit_condition() { return _why_did_we_stop; }
    
        /// set the maximum damping parameter
        virtual void set_max_damping_parameter(const double & d) { _max_damping_parameter = d; }
        /// set the minimum damping parameter
        virtual void set_min_damping_parameter(const double & d) { _min_damping_parameter = d; }
        /// set maximum number of consecutive failures 
        virtual void set_max_number_failures(const int & f) { _max_number_failures = f; }
        /// set the worst condition number for the matrix A
        virtual void set_worst_tolerated_condition_number(const double & d) {  }
        /// Pass to the MERLIN_SVDSolver member the threshold.
        virtual void set_svd_threshold(const double &pass_along_d) { }
        ///
        virtual void set_max_planck_function(double &val) { return limits_for_inversion.set_max_planck_function(val) ; }
        ///
        virtual void set_max_planck_function_gradient(double &val)  { return limits_for_inversion.set_max_planck_function_gradient(val) ; }
        /// Printer method, required by MERLIN_Object.
        virtual void dump( ostream &strm ) const {
            strm<<"Type of Inverter: Stepwise Regression"<<endl;
        };
    
    
    };

} // end namespace CSAC
#endif // MERLIN_InversionStepWiseRegression_h_


