/* ============================================================================
 * Language: C++ 
 * File: MERLIN_LScoupling.cc
 * Package Name: Milne Eddington gRid Linear Inversion Network (MERLIN).
 * Sub-package: Fundamental Inversion Code (FICO).
 *
 * Created by Jose Garcia <jgarcia@ucar.edu> on 11/11/05.
 *  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact University Corporation for Atmospheric Research at
 * 3080 Center Green Drive, Boulder, CO 80301

 * (c) COPYRIGHT University Corporation for Atmostpheric Research 2004-2006
 * Please read the full copyright statement in the file COPYRIGHT_UCAR.
 *
 =========================================================================== */

#include <string.h>
#include <stdlib.h>

#include <iostream>
using std::cout ;
using std::endl ;

#include <map>
using std::map ;

#include <string>
using std::string ;


#include "MERLIN_LScoupling.h"
using CSAC::MERLIN_LScoupling;


MERLIN_LScoupling::MERLIN_LScoupling( const char *coupling )
{
    /********************************************************
     * comes in as SLJ, can have spaces in it
     * S is an unsigned integer that is actually 2S+1
     * L is an enumeration, a single letter, save the letter and the enum
     * J is a double, but doesn't have to have a decimal. Save the integer
     * part, the decimal part, and the whole thing as a double.
     */
  
    // remove the leading and trailing blanks
    // if the first character is not a digit, throw an exception
    // read up to the first non digit
    // read the non digit, should be one of the enumeration characters
    // if next character not digit and not decimal then throw exception
    // if not decimal then read up to decimal or end of str, int part of J
    // if reached decimal then read past decimal to end of str, dec part of J
    // if found a non digit in J then throw an exception
  
    if( !coupling )
    {
        throw string( "null coupling string" ) ;
    }
  
    unsigned int len = strlen( coupling ) ;
  
    if( len < 3 )
    {
        throw string( "incorrect syntax in coupling, length is less than 3" ) ;
    }
  
    string c = coupling ;
  
    // remove leading blanks
    unsigned int index = 0 ;
    while( coupling[index] == ' ' && index < len )
    {
        index++ ;
    }
    if( index == len-1 )
    {
        throw string( "empty coupling string" ) ;
    }
  
    // pull out S (any digits before L enum character)
    unsigned int s_start = index ;
    if( !isdigit( coupling[index] ) )
    {
        throw string( "incorrect syntax in coupling, s contains non-digit" ) ;
    }
    while( isdigit( coupling[index] ) && index < len )
    {
        index++ ;
    }
    unsigned int s_end = index ;
    string s = c.substr( s_start, s_end-s_start ) ;
    _Sval = atoi( s.c_str() ) ;
    if( _Sval == 0 )
    {
        throw string( "S must be greater than 0" ) ;
    }
    if( _Sval == 1 ) _Sval = 0 ;
    else _Sval = (_Sval - 1)/2 ;
  
  
    // next character is L
    _Lchar = coupling[index] ;
    set_L() ;
    index++ ;
  
    // rest of string is J, 1.1, .1, 1
    unsigned int jint_start = index ;
    while( isdigit( coupling[index] ) && index < len )
    {
        index++ ;
    }
    unsigned int jint_end = index ;
    if( jint_start == jint_end )
    {
        if( index == len-1 )
        {
	    throw string( "no J specified in coupling" ) ;
        }
        if( coupling[index] != '.' )
        {
	    throw string( "incorrect syntax for j" ) ;
        }
        _Jint = 0 ;
    }
  
    unsigned int jdec_start = 0 ;
    unsigned int jdec_end = 0 ;
    // skip over the decimal
    if( index != len )
    {
        if( coupling[index] != '.' )
        {
	    throw string( "incorrect syntax for j, not at end and no dec" ) ;
        }
        index++ ;
    
        jdec_start = index ;
        while( isdigit( coupling[index] ) && index < len )
        {
	    index++ ;
        }
        jdec_end = index ;
    }
  
    if( !jdec_start )
    {
        string j = c.substr( jint_start, jint_end-jint_start ) ;
        _Jdec = 0 ;
        _Jint = atoi( j.c_str() ) ;
        _Jval = (double)_Jint ;
    }
    else
    {
        string j = c.substr( jint_start, jdec_end-jint_start ) ;
        _Jval = atof( j.c_str() ) ;
        string jint = c.substr( jint_start, jint_end-jint_start ) ;
        _Jint = atoi( jint.c_str() ) ;
        string jdec = c.substr( jdec_start, jdec_end-jdec_start ) ;
        _Jdec = atoi( jdec.c_str() ) ;
    }
}

void
MERLIN_LScoupling::set_L()
{
    switch( _Lchar )
    {
        case 'S':
            {
                _Lval = MERLIN_LScoupling::S ;
            }
	    break ;
        case 'P':
            {
                _Lval = MERLIN_LScoupling::P ;
            }
	    break ;
        case 'D':
            {
                _Lval = MERLIN_LScoupling::D ;
            }
	    break ;
        case 'F':
            {
                _Lval = MERLIN_LScoupling::F ;
            }
	    break ;
        case 'G':
            {
                _Lval = MERLIN_LScoupling::G ;
            }
	    break ;
        case 'H':
            {
                _Lval = MERLIN_LScoupling::H ;
            }
	    break ;
        case 'I':
            {
                _Lval = MERLIN_LScoupling::I ;
            }
	    break ;
        case 'K':
            {
                _Lval = MERLIN_LScoupling::K ;
            }
	    break ;
        case 'L':
            {
                _Lval = MERLIN_LScoupling::L ;
            }
	    break ;
        case 'M':
            {
                _Lval = MERLIN_LScoupling::M ;
            }
	    break ;
        case 'N':
            {
                _Lval = MERLIN_LScoupling::N ;
            }
	    break ;
        case 'O':
            {
                _Lval = MERLIN_LScoupling::O ;
            }
	    break ;
        case 'Q':
            {
                _Lval = MERLIN_LScoupling::Q ;
            }
	    break ;
        default:
            {
                throw string( "L is not valid" ) ;
            }
	    break ;
    }
}






