/* ============================================================================
 * Language: C++ 
 * File: MERLIN_Profile.cc
 * Package Name: Milne Eddington gRid Linear Inversion Network (MERLIN).
 * Sub-package: Fundamental Inversion Code (FICO).
 *
 * Created by Jose Garcia <jgarcia@ucar.edu> on 11/11/05.
 *  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact University Corporation for Atmospheric Research at
 * 3080 Center Green Drive, Boulder, CO 80301

 * (c) COPYRIGHT University Corporation for Atmostpheric Research 2004-2006
 * Please read the full copyright statement in the file COPYRIGHT_UCAR.
 *
 =========================================================================== */

#include <fstream>
using std::ifstream;

#include <iostream>
using std::endl;
using std::ostream;

#include <sstream>
using std::stringstream;

#include "MERLIN_Profile.h"
using CSAC::MERLIN_Profile;

#include "MERLIN_PrintFormated.h"
using CSAC::MERLIN_PrintFormated;


void
MERLIN_Profile::load_profile(const string &filename)
{
  
    ifstream profile_file( filename.c_str() ) ;
    if( !profile_file )
    {
        string err = string("The profile file ")
        + filename
        + " does not exist or cannot be opened" ;
        throw err ;
    }
    if( profile_file.eof() )
    {
        string err = string("The profile file ")
        + filename + " is empty" ;
        throw err ;
    }
	
    bool done = false ;
    double wave, i, q, u, v = 1.1 ;
    while( !done )
    {
        string buffer ;
        stringstream ss ;
        getline( profile_file, buffer ) ;
        if( profile_file.eof() )
        {
            done = true ;
        }
        else
        {
            ss.str( buffer ) ;
            ss >> wave >> i >> q >> u >> v ;
            stokes.lambda.push_back( wave );
            stokes.I.push_back( i ) ;
            stokes.Q.push_back( q ) ;
            stokes.U.push_back( u ) ;
            stokes.V.push_back( v ) ;
        }
    }
}

void
MERLIN_Profile::dump( ostream &strm ) const
{
    strm.precision( 10 ) ;
    typedef vector< double >::const_iterator citer ;
    citer wavei = stokes.lambda.begin();
    citer ii = stokes.I.begin() ;
    citer qi = stokes.Q.begin() ;
    citer ui = stokes.U.begin() ;
    citer vi = stokes.V.begin() ;
    for( ii; ii != stokes.I.end(); wavei++,ii++,qi++,ui++,vi++ )
    {
        MERLIN_PrintFormated::block_size = 10; 
        MERLIN_PrintFormated::print_blocked(strm,"lambda = "); 
        MERLIN_PrintFormated::block_size = 14; 
        MERLIN_PrintFormated::print_blocked(strm,*wavei); 
        MERLIN_PrintFormated::block_size = 6; 
        MERLIN_PrintFormated::print_blocked(strm,"I = "); 
        MERLIN_PrintFormated::block_size = 14; 
        MERLIN_PrintFormated::print_blocked(strm,*ii);
        MERLIN_PrintFormated::block_size = 6; 
        MERLIN_PrintFormated::print_blocked(strm,"Q = "); 
        MERLIN_PrintFormated::block_size = 14; 
        MERLIN_PrintFormated::print_blocked(strm,*qi);
        MERLIN_PrintFormated::block_size = 6; 
        MERLIN_PrintFormated::print_blocked(strm,"U = ");
        MERLIN_PrintFormated::block_size = 14; 
        MERLIN_PrintFormated::print_blocked(strm,*ui);
        MERLIN_PrintFormated::block_size = 6; 
        MERLIN_PrintFormated::print_blocked(strm,"V = "); 
        MERLIN_PrintFormated::block_size = 14; 
        MERLIN_PrintFormated::print_blocked(strm,*vi);
        strm << endl ;
    }
}

void
MERLIN_Profile::resize(const int &number_of_free_parameters, const int &number_of_wavelenghts)
{
    //Reserve space for the vector containing the wavelengths for which I, Q, U and V are to be computed
    stokes.lambda.resize(number_of_wavelenghts, 0.0);
    // Prepare space for I, Q, U and V
    stokes.I.resize(number_of_wavelenghts, 0.0);
    stokes.Q.resize(number_of_wavelenghts, 0.0);
    stokes.U.resize(number_of_wavelenghts, 0.0);
    stokes.V.resize(number_of_wavelenghts, 0.0);
  
    // Reserve space to annotate the derivatives order and set order
    stokes_derivatives.order.resize(number_of_free_parameters,""); 
 
    // Prepare derivatives of the I, Q, U and V w.r.t free parameters
    vector <double> dummy_vector;
    dummy_vector.resize(number_of_wavelenghts, 0.0);
    stokes_derivatives.I_dxdy.resize(number_of_free_parameters, dummy_vector);
    stokes_derivatives.Q_dxdy.resize(number_of_free_parameters, dummy_vector);
    stokes_derivatives.U_dxdy.resize(number_of_free_parameters, dummy_vector);
    stokes_derivatives.V_dxdy.resize(number_of_free_parameters, dummy_vector);  
}

bool 
MERLIN_Profile::match_size(const int &number_of_free_parameters, const int &number_of_wavelenghts)
{
    if ( stokes.lambda.size() != number_of_wavelenghts ||
         stokes.I.size()      != number_of_wavelenghts || 
         stokes.Q.size()      != number_of_wavelenghts ||
         stokes.U.size()      != number_of_wavelenghts ||
         stokes.V.size()      != number_of_wavelenghts )
        return false;
  
    if ( stokes_derivatives.order.size()  != number_of_free_parameters  ||
         stokes_derivatives.I_dxdy.size() != number_of_free_parameters  ||
         stokes_derivatives.Q_dxdy.size() != number_of_free_parameters  ||
         stokes_derivatives.U_dxdy.size() != number_of_free_parameters  ||
         stokes_derivatives.V_dxdy.size() != number_of_free_parameters  )
        return false;
  
    for (int i = 0; i < number_of_free_parameters; i++)
    {
        if ( stokes_derivatives.I_dxdy[i].size() != number_of_wavelenghts ||
             stokes_derivatives.Q_dxdy[i].size() != number_of_wavelenghts ||
             stokes_derivatives.U_dxdy[i].size() != number_of_wavelenghts ||
             stokes_derivatives.V_dxdy[i].size() != number_of_wavelenghts )
            return false;
    }
  
    return true;
}

