/* ============================================================================
 * Language: C++ 
 * File: MERLIN_StrayLight.h
 * Package Name: Milne Eddington gRid Linear Inversion Network (MERLIN).
 * Sub-package: Fundamental Inversion Code (FICO).
 *
 * Created by Jose Garcia <jgarcia@ucar.edu> on 11/11/05.
 *  
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact University Corporation for Atmospheric Research at
 * 3080 Center Green Drive, Boulder, CO 80301

 * (c) COPYRIGHT University Corporation for Atmostpheric Research 2004-2006
 * Please read the full copyright statement in the file COPYRIGHT_UCAR.
 *
 =========================================================================== */

#ifndef MERLIN_StrayLight_h_
#define MERLIN_StrayLight_h_ 1

#include <vector>
using std::vector;
#include <string>
using std::string;

#include "MERLIN.h"
#include "MERLIN_Object.h"
using CSAC::MERLIN_Object;
#include "MERLIN_Range.h"
using CSAC::MERLIN_Range;


namespace CSAC
{
  

    /// \brief This class encapsulates the concept of Stray Light to be added to Stokes profiles
    /// synthetized by Milne-Eddington. 
    ///
    /// As part of the GIMME framework, this class supports the developers in the sense that it provides
    ///  a mechanism by which a vector of numbers representing the stray light, can be:
    ///  -# transparently shifted according to the shift theorem.
    ///  -# derived with respect to the shift.
    ///  -# accessed as if the object was a vector in itself.
    ///  -# constraint according to an specific range of operation. \n\n
    /// Consider the following stray light signal\n
    /// \image html stray_light.png
    /// @see set_ranges. 

    class MERLIN_StrayLight : public MERLIN_Object
    {	
        /// \brief Object internal size.
        ///
        /// This variable tells us the size of the buffers 
        /// _unshifted_light, _unshifted_light_in_fourier_space and _shifted_light
        int _size;
        /// \brief This is where the original unshifted light given is held.
        double *_unshifted_light;
        /// \brief This is where the continuum is held
        double _continuum;
        /// \brief This is where the original unshifted light given is held as it is in Fourier space.
        double *_unshifted_light_in_fourier_space;
        /// \brief This is where the shifted light in real space is held.
        ///
        /// The shifted light is the result of the shift theorem applied to _unshifted light
        /// The amount of shift is determined by the current by the _shift value
        double *_shifted_light;
        /// \brief Derivative of the signal with respect to the shift.
        double *_derivative_shifted_light;
        /// \brief Buffer to store the precomputed value for the derivative when the shift is zero
        double *_derivative_shifted_light_with_shift_zero;
        /// \brief Current shift of the light in this object.
        ///
        /// As the light stored in this object is shifted by a value
        /// according to the shift theorem, this variable holds the
        /// value for the current shift. @see shift_light
        double _shift;
        /// Is this object loaded with data?
        bool _set;
        ///
        vector<CSAC::MERLIN_Range> _ranges;
    public:
		
        /// \brief Constructor. 
        ///
        /// After creation, the object has all internal buffers are set to zero, 
        /// its size is set to zero and the flag _set is false.
        MERLIN_StrayLight() : 
            _shifted_light(0), 
            _derivative_shifted_light(0),
            _derivative_shifted_light_with_shift_zero(0),
            _unshifted_light(0),
            _continuum(0),
            _unshifted_light_in_fourier_space(0), 
            _size(0), 
            _shift(0),
            _set(false)
        {	
        }
  
	
        /// \brief Copy Constructor.
        ///
        /// Since this class allocates buffers of data, the copy constructor is implemented
        /// to ensure that brand new objects of this class created as copy of previous objects
        /// have their own buffers, etc.
        MERLIN_StrayLight(const MERLIN_StrayLight &rhs);
	
        /// \brief Destructor.
        ///
        /// Ensures that the object properly disposes its buffers. @see void clean()
        virtual ~MERLIN_StrayLight();
	
        /// @name Accessors
        /// Provides a mechanism to access the stray light  as it is shifted by the current _shift value
        /// as well as to query the status of a StrayLight object.
        //@{
        /// \brief Informs you if the object is already set.
        ///
        /// If you already called the method set_stray_light, then the object is loaded and 
        /// a call to set() will return true, otherwise the object is empty (as it is as soon it is constructed)
        /// and set() will return false.
        /// @see set_stray_light.
        /// 
        bool set() const {return _set;}
        /// \brief Returns the size of this object.
        ///
        /// This method is used mostly to know how to access the shifted light
        /// using the overloaded operator []. For example
        /// @see _shift.
        int size() {return _size;}
        /// \brief Returns the size of this constant object.
        ///
        /// This method is used mostly to know how to access the shifted light
        /// using the overloaded operator []. For example
        /// @see _shift.
        int size() const {return _size;}
        /// \brief overloaded [] operator for constant objects.
        ///
        /// if the position requested is out of bounds, then an error message as a string is thrown.
        /// @see _size.
        double& operator[](int i) const throw (string);
        /// \brief overloaded [] operator.
        ///
        /// if the position requested is out of bounds, then an error message as a string is thrown.
        /// @see _size.
        double& operator[](int i) throw (string);
        //@}
	
        /// @name Modifiers
        /// Provides a mechanism to modify a stray light object allowing the programmer to set the value of
        /// the stray light signal, shift the light, and load a constrained portion of the signal defined 
        /// by a set a set of ranges.  
        //@{
        /// \brief set the contents of this object. 
        ///
        /// If you a signal already representing the stray light, then you can load it
        /// into an object of this class by using this method. Please notice, you current
        /// vector holding the signal MUST BE A POWER OF TWO in size, otherwise you will get an exception.
        void set_stray_light(const vector<double> &) throw (string);
        /// \brief Shift the signal by some pixels.
        ///
        /// This method applies the shift theorem to a signal already loaded and have the object ready for access.
        /// If by mistake you forget to properly set you object before shiftting the signal, you will get an exception.
        void shift_light(const double &) throw (string);
        /// \brief Copy contents of another StrayLight object.
        ///
        /// Since objects of this class used dynamically allocated memory, it is imperative that
        /// this operator is overloaded to proper release any currently allocated memory and update contents.
        MERLIN_StrayLight& operator=(const MERLIN_StrayLight&);
        /// \brief Cleans internal buffers.
        ///
        /// If the internal buffers are set, the clean releases the memory, and resets all 
        /// internal buffers to zero. The internal size is also set to zero and the flag _set 
        /// is assigned false.
        void clean();
        //@}
	
        /// @name Constrained accessors.
        /// If you are interested in just one portion of the signal, this methods 
        /// will give you a mechanism to access such portion.
        //@{
        /// \brief loads just a portion of the current signal help on this object as defined by ranges.
        ///
        /// @see set_ranges. 
        void load_constrained_stray_light_vector(vector<double> &);
        ///  \brief loads just a portion of the wavelenghts held on this objects as defined by ranges.
        ///
        /// @see set_ranges.
        void load_constrained_stray_light_vector_wavelength(vector<double> &);
        /// 
        void load_constrained_derivative_wrt_current_shift(vector<double> &);
        /// \brief Set a constrained range for the signal held in the object
        void set_ranges (const vector<CSAC::MERLIN_Range>&) throw (string);
        //@} 
	
        /// \brief Dump the contents of this object to a stream.
        ///
        /// When debugging or in order to visualize the contents of an object
        /// you can simply invoke
        /** 
            \code
            #include "StrayLight.h"
            ...
            StrayLight my_stray_light;
            ...
            cout<<my_stray_light;
            \endcode
        */
        virtual void dump( ostream &strm ) const ;
	
        /// \brief Derivative of the of the Stray Light with respect to the shift.
        ///
        /// Let \f$S\f$ be the stray light and let \f$Sf\f$ be the stray light on Fourier space.
        /// If there is a doppler shift on the signal sampling,
        /// we can obtained such shifted signal by using the shift therorem. 
        /// According to the shift theorem, for a shift \f$\delta\f$, the shift of the 
        /// stray light can be obtained by \f$FFT^{-1}(Sf_\delta)\f$, which is defined by: \n\n
        /// \f$ Sf_\delta = exp(\frac{2\pi k\delta}{n_p\Delta\lambda_p}i)Sf \f$ \n\n
        /// in this case, the derivative with respect of the shift is equal to 
        /// \f$FFT^{-1}(\frac{\partial Sf_\delta}{\partial \delta})\f$ defined by:\n\n
        /// \f$ \frac{\partial Sf_\delta}{\partial \delta} = (\frac{2\pi k}{n_p\Delta\lambda_p}i) Sf_\delta\f$
        void derivative_wrt_current_shift(vector<double> &) throw(string);
        ///
        double get_continuum() {
            return _continuum;
        }
	
    }; // end class StrayLight

}; // end namespace CSAC

#endif // StrayLight_h_
