Subroutine Invert_Field(nlambda, SI, SQ, SU, SV, StrayProf, &
     continuum1, continuum2, core1, core2, Field, Incli, Vel, Ffactor, pixel)
! may use PixShift instead of Vel, if you want the integer shift
!
! Inputs:
!
!   ANNfile: String (length=256) array with the full path and filename to the
!             binary neural network file. The first element has the path
!             to the ANN for field strength. The second element has the 
!             path to the ANN for inclination.
!   nlambda: (Integer), number of wavelengths in the spectral profile
!   SI, SQ, SU, SV: Real vectors of dimension [nlambda]. Observed profiles
!   StrayProf: (Real vector of dimension nlambda) Stray light profile
!   continuum1, continuum2: (Integers) Pixel positions marking the limits of
!                a continuum window in the profile
!   core1, core2: (Integers) Pixel positions marking a window that encloses
!                the minimum of the 6301.5 line
!
! Outputs:
!
!   Field: (Real) Magnetic field strength from the ANN inversion
!   Incli: (Real) Magnetic field inclination from the ANN inversion
!   Vel: (Real) Velocity from the ANN inversion (average of magnetic 
!             and surroundings) in km/s. Positive -> Redshift
!   Ffactor: (Real) Filling factor of magnetic component (between 0 and 1)
!
  Implicit None
  Real, Parameter :: CC=3.e5 ! km/s
  Real, Parameter :: mA_per_pixel=21.53 ! Hinode
  Integer :: continuum1, continuum2, core1, core2, nlambda, fileunit
  Real, Dimension(:), Allocatable :: X, Y, Xp
  Real, Dimension(nlambda) :: SI, SQ, SU, SV
  Real, Dimension(:,:), Allocatable, Save :: Profile
  Real, Dimension(nlambda) :: StrayProf
  Real :: Field, Incli, Vel, Ffactor
  Real :: icross, Chisq, MinChisq
  Character (Len=256), Dimension(4) :: ANNfile
  Integer :: istep, i, index, Pixshift, inet, icrossref, pixel
  Integer, Dimension(1) :: imin, imax
  Logical :: FlipInc
! Variables set the first time and saved for later calls
  Logical, Save :: FirstTime=.True.
  Integer, Save :: nlayers, ninputs, noutputs, nmaxperlayer
  Integer, Dimension(:), Allocatable, Save :: nperlayer
  Real, Dimension(:,:, :, :), Allocatable, Save :: W
  Real, Dimension(:,:, :), Allocatable, Save :: Beta
  Real, Dimension(:, :, :), Allocatable, Save :: W0
  Real, Dimension(:, :), Allocatable, Save :: Beta0
  Real, Dimension(:,:), Allocatable, Save :: tmp
  Real, Dimension(:,:), Allocatable, Save :: Xnorm, Xmean
  Real, Dimension(:,:), Allocatable, Save :: Ynorm, Ymean
  Integer, Dimension(:), Allocatable, Save :: nonlin, IndexProf
!

!print *,'Pixel given is ',pixel
!print *,'SI= ',SI
!print *,'SQ= ',SQ

ANNfile(1)='/export/jgarcia/bin/merlin/etc/Field.net.swap'
ANNfile(2)='/export/jgarcia/bin/merlin/etc/Incli.net.swap'
ANNfile(3)='/export/jgarcia/bin/merlin/etc/Vel.net.swap'
ANNfile(4)='/export/jgarcia/bin/merlin/etc/Ffactor.net.swap'
  If (FirstTime) then
! Allocate memory for Profiles
     Allocate (Profile(4, nlambda))
! Hardwired values. These should be included in the ANNfile at some point
     nlayers=4
     ninputs=80
     noutputs=5
     nmaxperlayer=80

     Allocate (nonlin(nlayers))
     nonlin(1:nlayers)=1

     Allocate (nperlayer(0:nlayers))
     nperlayer(0)=ninputs
     nperlayer(1:nlayers)=20
! Specify which wavelengths to pick (20 for each line, ninputs=20*4)
     Allocate (IndexProf(20))
     istep=2
     Do i=1, 10
        IndexProf(i)=20+i*istep ! Hardwired value here
        IndexProf(i+10)=65+i*istep ! Hardwired value here
     End do
! Allocate arrays
     Allocate (W0(nlayers, nmaxperlayer, nmaxperlayer)) 
     Allocate (Beta0(nlayers, nmaxperlayer)) 
     Allocate (W(4, nlayers, nmaxperlayer, nmaxperlayer)) 
     Allocate (Beta(4, nlayers, nmaxperlayer)) 
     Allocate (Xnorm(4, ninputs))
     Allocate (Ynorm(4, noutputs))
     Allocate (Xmean(4, ninputs))
     Allocate (Ymean(4, noutputs))
     Allocate (tmp(0:nlayers, nmaxperlayer))
! Read from files
     Do inet=1, 4
        Call Open_file_unform(fileunit, ANNfile(inet))
        Read (fileunit) W(inet,:,:,:)
        Read (fileunit) Beta(inet,:,:)
        Read (fileunit) Xnorm(inet,:), Xmean(inet,:)
        Read (fileunit) Ynorm(inet,:), Ymean(inet,:)
        Close (fileunit)
     End do
     FirstTime=.False.
  Endif
!
  Profile(1,1:nlambda)=SI(1:nlambda)
  Profile(2,1:nlambda)=SQ(1:nlambda)
  Profile(3,1:nlambda)=SU(1:nlambda)
  Profile(4,1:nlambda)=SV(1:nlambda)
!
! Normalize the quiet Sun profile to its continuum intensity
!
  Profile(1:4, 1:nlambda)=Profile(1:4, 1:nlambda)/ &
       (Sum(Profile(1, continuum1:continuum2))/(continuum2-continuum1+1))
  StrayProf(1:nlambda)=StrayProf(1:nlambda)/ &
       (Sum(StrayProf(continuum1:continuum2))/(continuum2-continuum1+1))
!
! Find center of symmetry of quiet Sun profile around pixel 0.5*(core1+core2)
!
  icrossref=.5*(core1+core2)
  imin=MinLoc(StrayProf(core1:core2))+core1-1 ! Minimum of Stokes I
  MinChisq=1.e10
  Do index=Max(imin(1)-15,21), &
       Min(imin(1)+15, nlambda-21) ! Find symmetry center of Stokes I
     Chisq=Sum(Abs(StrayProf(index-20:index) - &
          StrayProf(index+20:index:-1)))
     If (Chisq .lt. MinChisq) then
        MinChisq=Chisq
        icross=index
     End if
  End do
! Apply shift  
  index=icross-icrossref
  If (index .gt. 0) then
     StrayProf(1:nlambda-index) = StrayProf(1+index:nlambda)
  Else If (index .lt. 0) then
     StrayProf(1-index:nlambda) = StrayProf(1:nlambda+index)
  End if
!
! Find center of symmetry of observed profile around pixel 0.5*(core1+core2)
!
  icrossref=.5*(core1+core2)
  imin=MinLoc(Profile(1, core1:core2))+core1-1 ! Minimum of Stokes I
  MinChisq=1.e10
  Do index=Max(imin(1)-15,21), &
       Min(imin(1)+15, nlambda-21) ! Find symmetry center of Stokes I
     Chisq=Sum(Abs(Profile(1, index-20:index) - &
          Profile(1, index+20:index:-1)))
     If (Chisq .lt. MinChisq) then
        MinChisq=Chisq
        icross=index
     End if
  End do
! Apply shift  
  index=icross-icrossref
  If (index .gt. 0) then
     Profile(1:4, 1:nlambda-index) = Profile(1:4, 1+index:nlambda)
  Else If (index .lt. 0) then
     Profile(1:4, 1-index:nlambda) = Profile(1:4, 1:nlambda+index)
  End if
  Pixshift=-index ! By how much has the profile been redshifted (integer)
!
! Change polarity if necessary
!
  FlipInc=.False.
  If (Sum(Profile(4, icrossref-10:icrossref)) .gt. &
       Sum(Profile(4, icrossref:icrossref+10))) then ! Reverse
     Profile(4,:)=-Profile(4,:) ! Set inclination between 90-180 deg
     FlipInc=.True.
  End if
!
! Extract ANN inputs from observed profile
!
  Allocate (X(ninputs))
  Allocate (Xp(ninputs))
  Allocate (Y(noutputs))
  Do i=1, ninputs/4
     X(i)=Sum(Profile(1, IndexProf(i): &
          IndexProf(i)+istep-1))/istep
     X(ninputs/4+i)=Sqrt( &
          (Sum(Profile(2, IndexProf(i): &
          IndexProf(i)+istep-1))/istep)**2 + &
          (Sum(Profile(3, IndexProf(i): &
          IndexProf(i)+istep-1))/istep)**2)
     X(2*ninputs/4+i)=Sum(Profile(4, &
          IndexProf(i):IndexProf(i)+istep-1))/istep
     X(3*ninputs/4+i)=Sum(StrayProf(IndexProf(i): &
          IndexProf(i)+istep-1))/istep
  End do
!
! Pre-process inputs with linear transformation
!
  Do i=1, ninputs
     Xp(i)=(X(i)-Xmean(1,i))/Xnorm(1,i)
  End do
!
! Call ANN to obtain the model parameter
!
! First the field strength
  W0=W(1,:,:,:)
  Beta0=Beta(1,:,:)
  Call ANN_forward(W0, Beta0, nonlin, Xp, Y, nlayers,  &
       nmaxperlayer, nperlayer, ninputs, noutputs, tmp)
!
! Post-process output with linear transformation
!
  Field=Y(1)*Ynorm(1,1)+Ymean(1,1)

 !print *, 'Field is ', Field

! Now for the inclination
  Do i=1, ninputs
     Xp(i)=(X(i)-Xmean(2,i))/Xnorm(2,i)
  End do
  W0=W(2,:,:,:)
  Beta0=Beta(2,:,:)
  Call ANN_forward(W0, Beta0, nonlin, Xp, Y, nlayers,  &
       nmaxperlayer, nperlayer, ninputs, noutputs, tmp)
  Incli=Y(1)*Ynorm(2,1)+Ymean(2,1)
! Flip inclination if necessary
  If (FlipInc) Incli=180.-Incli
!
! Now for the velocity
  Do i=1, ninputs
     Xp(i)=(X(i)-Xmean(3,i))/Xnorm(3,i)
  End do
  W0=W(3,:,:,:)
  Beta0=Beta(3,:,:)
  Call ANN_forward(W0, Beta0, nonlin, Xp, Y, nlayers,  &
       nmaxperlayer, nperlayer, ninputs, noutputs, tmp)
  Vel=Y(1)*Ynorm(3,1)+Ymean(3,1)
  Vel=Vel-Pixshift*mA_per_pixel*1.E-3/6302.*CC
!
! Now for the filling factor
  Do i=1, ninputs
     Xp(i)=(X(i)-Xmean(4,i))/Xnorm(4,i)
  End do
  W0=W(4,:,:,:)
  Beta0=Beta(4,:,:)
  Call ANN_forward(W0, Beta0, nonlin, Xp, Y, nlayers,  &
       nmaxperlayer, nperlayer, ninputs, noutputs, tmp)
  Ffactor=Y(1)*Ynorm(4,1)+Ymean(4,1)
!
  Return

End Subroutine Invert_Field


 
